part of jaguar.generator.parser.route;

class ParsedRouteWrapper extends NamedElement
    with ChainFunction, MethodReturnTypeMixin {
  String get name => type.name;

  String get libraryName => type.libraryName;

  DartTypeWrap get type => method.returnTypeWithoutFuture;

  final ClassElementWrap wrapped;

  /// The function or method element
  final MethodElementWrap method;

  /// Inputs declared on the interceptor
  final List<ParsedInput> inputs = [];

  ParsedRouteWrapper(this.wrapped, this.method) {
    inputs.addAll(_detectInputs());
  }

  bool get canHaveQueryParams => false;

  static int getNumDefaultInputs(bool httpReq) => httpReq ? 1 : 0;

  @override
  int get numDefaultInputs => getNumDefaultInputs(needsHttpRequest);

  //Detects inputs
  List<ParsedInput> _detectInputs() {
    final List<ParsedInput> inputs =
    ParsedInput.detectInputs(method, method.parameters, numDefaultInputs);

    final int numMethodInps = numDefaultInputs + inputs.length;
    bool hasFinished = false;

    //Detect inputs on parameters
    for (int idx = numDefaultInputs;
    idx < method.requiredParameters.length;
    idx++) {
      final ParameterElementWrap param = method.requiredParameters[idx];

      ParsedInput input = ParsedInput.detectOnParam(param);

      if (idx < numMethodInps) {
        if (input is ParsedInput) {
          final except = new InputException(
              'Input for this method is already specified on method!');
          except.param = param.name;
          throw except;
        }
        continue;
      }

      if (input is ParsedInput) {
        if (hasFinished) {
          final except = new InputException(
              'Inputs must be specified in consecutive params!');
          except.param = param.name;
          throw except;
        }
        inputs.add(input);
      } else {
        hasFinished = true;
      }
    }

    return inputs;
  }

  factory ParsedRouteWrapper.FromProperty(MethodElementWrap meth) {
    DartTypeWrap type = meth.returnTypeWithoutFuture;

    if (!type.isSubTypeOfNamedElement(kTypeRouteWrapper)) {
      throw new Exception("Route wrapper provider must return RouteWrapper!");
    }

    InterfaceTypeWrap interfaceType =
        type.clazz.getSubtypeOf(kTypeRouteWrapper);
    if (interfaceType.typeArguments.length != 1) {
      throw new Exception('Provide Interceptor in RouteWrapper!');
    }

    ClassElementWrap clazz = interfaceType.typeArguments.first.clazz;

    return new ParsedRouteWrapper(clazz, meth);
  }
}
