part of jaguar.generator.parser.route;

abstract class ParsedWrapper {
  ClassElementWrap get interceptorClazz;

  /// Pre interceptor info
  ParsedInterceptorFuncDef get pre;

  /// Post interceptor info
  ParsedInterceptorFuncDef get post;

  /// Detects interceptors on a given method or class
  static List<ParsedWrapper> detectWrappers(
      ClassElementWrap upper, WithMetadata element) {
    final ret = <ParsedWrapper>[];

    for (AnnotationElementWrap annot in element.metadata) {
      final instantiated = annot.instantiated;
      if (instantiated is ant.Wrap) {
        for (Symbol sym in instantiated.interceptorChain) {
          final wrap =
              new WrapperCreator.Make(upper, MirrorSystem.getName(sym));
          ret.add(wrap);
        }
      } else {
        final ClassElementWrap clazz = annot.ancestorClazz;

        if (clazz is! ClassElementWrap) {
          continue;
        }

        if (!clazz.isSubtypeOf(kTypeRouteWrapper)) {
          continue;
        }

        ret.add(new WrapperInstance.Make(annot));
      }
    }

    return ret;
  }
}

class WrapperCreator extends Object implements ParsedWrapper {
  /// Holds the constant value of the interceptor annotation
  final ParsedRouteWrapper routeWrapper;

  /// The class of the interceptor
  final ClassElementWrap interceptorClazz;

  /// Pre interceptor info
  final ParsedInterceptorFuncDef pre;

  /// Post interceptor info
  final ParsedInterceptorFuncDef post;

  WrapperCreator(this.interceptorClazz, this.routeWrapper, this.pre, this.post);

  factory WrapperCreator.Make(ClassElementWrap upper, String sym) {
    MethodElementWrap field = upper.methods
        .firstWhere((MethodElementWrap f) => f.name == sym, orElse: () => null);
    if (field == null) {
      throw new Exception("RouteWrapper provider $sym not found!");
    }
    final routeWrapper = new ParsedRouteWrapper.FromProperty(field);
    final ClassElementWrap clazz = routeWrapper.wrapped;
    ParsedInterceptorFuncDef pre;
    ParsedInterceptorFuncDef post;

    /// Find pre and post interceptors in class
    clazz.methods.forEach((MethodElementWrap method) {
      if (method.name == 'pre') {
        pre = new ParsedInterceptorFuncDefBuilder(method).func;
      } else if (method.name == 'post') {
        post = new ParsedInterceptorFuncDefBuilder(method).func;
      }
    });

    Map<String, bool> interceptorResultsUsed = {};

    pre?.inputs?.forEach((ParsedInput inp) {
      if (inp is ParsedInputInterceptor) {
        interceptorResultsUsed[inp.genName] = true;
      }
    });

    post?.inputs?.forEach((ParsedInput inp) {
      if (inp is ParsedInputInterceptor) {
        interceptorResultsUsed[inp.genName] = true;
      }
    });

    return new WrapperCreator(clazz, routeWrapper, pre, post);
  }

  /// Return type of pre step
  DartTypeWrap get returnType => pre?.returnType;

  /// Flattened return type of pre step
  DartTypeWrap get returnsFutureFlattened => pre?.returnsFutureFlattened;

  /// Constructor of the interceptor
  ConstructorElementWrap get constructor => interceptorClazz.unnamedConstructor;
}
