part of jaguar.generator.parser.route;

class WrapperInstance extends Object implements ParsedWrapper {
  final AnnotationElementWrap annotation;

  DartTypeWrap get wrapperType =>
      new DartTypeWrap(annotation.constantValue.type);

  final String id;

  /// The class of the interceptor
  final ClassElementWrap interceptorClazz;

  /// Pre interceptor info
  final ParsedInterceptorFuncDef pre;

  /// Post interceptor info
  final ParsedInterceptorFuncDef post;

  WrapperInstance(
      this.annotation, this.id, this.interceptorClazz, this.pre, this.post);

  factory WrapperInstance.Make(AnnotationElementWrap annot) {
    String id;
    ClassElementWrap interceptorClazz;

    {
      DartObject wrapperConstVal = annot.constantValue;

      final wrapperType = new DartTypeWrap(wrapperConstVal.type);
      id = wrapperConstVal.getField('id')?.toStringValue();

      InterfaceTypeWrap interfaceType =
          wrapperType.clazz.getSubtypeOf(kTypeRouteWrapper);
      if (interfaceType.typeArguments.length != 1) {
        throw new Exception('Provide Interceptor in RouteWrapper!');
      }

      interceptorClazz = interfaceType.typeArguments.first.clazz;
    }

    ParsedInterceptorFuncDef pre;
    ParsedInterceptorFuncDef post;

    /// Find pre and post interceptors in class
    interceptorClazz.methods.forEach((MethodElementWrap method) {
      if (method.name == 'pre') {
        pre = new ParsedInterceptorFuncDefBuilder(method).func;
      } else if (method.name == 'post') {
        post = new ParsedInterceptorFuncDefBuilder(method).func;
      }
    });

    return new WrapperInstance(annot, id, interceptorClazz, pre, post);
  }

  /// Return type of pre step
  DartTypeWrap get returnType => pre?.returnType;

  /// Flattened return type of pre step
  DartTypeWrap get returnsFutureFlattened => pre?.returnsFutureFlattened;
}
