library jaguar_generator.toModels;

import 'package:jaguar_generator/parser/parser.dart';
import 'package:jaguar_generator/models/models.dart';
import 'package:source_gen_help/source_gen_help.dart';

import 'package:jaguar_generator/common/exceptions/exceptions.dart';

class ToModelUpper {
  ParsedUpper upper;

  final Upper _ret = new Upper();

  ToModelUpper(this.upper) {
    _perform();
  }

  Upper toModel() => _ret;

  void _perform() {
    _ret.baseName = upper.name;
    _ret.prefix = upper.path;
    _ret.emptyConstructor =
        upper.upper.unnamedConstructor.requiredParameters.length == 0;

    for (ParsedRoute route in upper.routes) {
      _ret.addMethod(new Method(route.method.name, route.method.prototype));

      Route newRoute = new Route();
      newRoute.instantiation = 'const ' + route.instantiationString;
      newRoute.prototype = route.prototype;
      newRoute.value = route.item;
      newRoute.isWebsocket = route.isWebSocket;
      newRoute.needsHttpRequest = route.needsHttpRequest;

      final interceptorNums = <String, int>{};

      for (ParsedWrapper interceptor in upper.interceptors) {
        final String name = interceptor.interceptorClazz.libraryName +
            '.' +
            interceptor.interceptorClazz.name;
        int number = 0;
        if (interceptorNums.containsKey(name)) {
          interceptorNums[name]++;
          number = interceptorNums[name];
        } else {
          interceptorNums[name] = 0;
        }
        try {
          newRoute.interceptors
              .add(makeInterceptor(route, interceptor, number));
        } on InterceptorWrapParamException catch (e) {
          e.on = upper.name;
          e.upper = upper.name;
          rethrow;
        }
      }

      for (ParsedWrapper interceptor in route.interceptors) {
        final String name = interceptor.interceptorClazz.libraryName +
            '.' +
            interceptor.interceptorClazz.name;
        int number = 0;
        if (interceptorNums.containsKey(name)) {
          interceptorNums[name]++;
          number = interceptorNums[name];
        } else {
          interceptorNums[name] = 0;
        }
        try {
          newRoute.interceptors
              .add(makeInterceptor(route, interceptor, number));
        } on InterceptorWrapParamException catch (e) {
          e.on = route.method.name;
          e.upper = upper.name;
          e.route = route.method.name;
          rethrow;
        }
      }

      _ret.routes.add(newRoute);

      {
        RouteMethod method = new RouteMethod();

        method.name = route.method.name;
        method.returnType = route.method.returnTypeWithoutFuture.displayName;
        method.returnsVoid = route.method.returnType.isVoid;
        method.returnsResponse = route.returnsResponse;
        if (!route.jaguarResponseType.isVoid) {
          method.jaguarResponseType = route.jaguarResponseType.name;
        } else {
          method.jaguarResponseType = 'dynamic';
        }
        method.isAsync = route.method.returnType.isAsync;

        List<Input> inputs = _makeInputs(route.inputs);

        List<PathParam> pathParams = <PathParam>[];
        List<QueryParam> queryParams = <QueryParam>[];

        for (ParameterElementWrap pathParam in route.nonInputParams) {
          if (!pathParam.type.isBuiltin) {
            pathParams.add(new PathParamNull());
          } else {
            pathParams.add(
                new PathParamKeyed(pathParam.name, _getStringTo(pathParam)));
          }
        }

        if (!route.areOptionalParamsPositional) {
          for (ParameterElementWrap queryParam in route.optionalParams) {
            queryParams.add(new QueryParam(queryParam.name,
                _getStringTo(queryParam), queryParam.toValueIfBuiltin));
          }
        }

        method.pathParams = pathParams;
        method.queryParams = queryParams;
        method.inputs = inputs;
        newRoute.method = method;
      }

      //Exceptions
      {
        List<ExceptionHandler> handlers = <ExceptionHandler>[];

        final func = (ParsedExceptionHandler except) => new ExceptionHandler(
            except.exceptionName,
            except.handlerName,
            except.instantiationString);

        upper.exceptions.map(func).forEach(handlers.add);
        route.exceptions.map(func).forEach(handlers.add);
        newRoute.exceptions = handlers;
      }

      {
        int respIndex = 0;
        for (Interceptor inter in newRoute.interceptors.reversed) {
          if (inter.post == null) {
            continue;
          }

          for (Input inp in inter.post.inputs) {
            if (inp is InputRouteResponse) {
              inp.respIndex = respIndex;
            }
          }

          if (inter.post.returnsResponse) {
            respIndex++;
            inter.post.respIndex = respIndex;
          }
        }
        newRoute.respIndex = respIndex;
      }
    }

    for (ParsedGroup group in upper.groups) {
      Group newGroup = new Group();

      newGroup.name = group.name;
      newGroup.path = group.group.path;
      newGroup.type = group.type.displayName;

      _ret.groups.add(newGroup);
    }
  }

  //Creates interceptor model for given parsed interceptor
  Interceptor makeInterceptor(
      ParsedRoute route, ParsedWrapper wrapper, int number) {
    Interceptor newInterceptor = new Interceptor();

    newInterceptor.name = wrapper.interceptorClazz.name;
    newInterceptor.number = number;

    //Make constructor
    if (wrapper is WrapperCreator) {
      final creator =
          new InterceptorCreatorMethod(wrapper.routeWrapper.method.name);
      newInterceptor.creator = creator;

      creator.needsHttpRequest = wrapper.routeWrapper.needsHttpRequest;
      creator.isAsync = wrapper.routeWrapper.returnType.isAsync;
      creator.inputs = _makeInputs(wrapper.routeWrapper.inputs);
    } else if (wrapper is WrapperInstance) {
      final creator = new InterceptorCreatorInstance(
          wrapper.annotation.instantiationString);
      newInterceptor.creator = creator;
    }

    //Make pre
    if (wrapper.pre is ParsedInterceptorFuncDef) {
      InterceptorPre pre = new InterceptorPre();
      pre.needsHttpRequest = wrapper.pre.needsHttpRequest;
      pre.returnType = wrapper.pre.returnsFutureFlattened.displayName;
      pre.isAsync = wrapper.pre.returnType.isAsync;
      pre.isVoid = wrapper.pre.returnType.isVoid;
      pre.inputs = _makeInputs(wrapper.pre.inputs);
      newInterceptor.pre = pre;
    }

    //Make post
    if (wrapper.post is ParsedInterceptorFuncDef) {
      InterceptorPost post = new InterceptorPost();
      post.needsHttpRequest = wrapper.post.needsHttpRequest;
      post.returnsResponse = wrapper.post.returnsResponse;
      post.isAsync = wrapper.post.returnType.isAsync;
      post.inputs = _makeInputs(wrapper.post.inputs);
      if (wrapper.post.returnsResponse) {
        post.jaguarResponseType = wrapper.post.jaguarResponseType.name;
      }
      newInterceptor.post = post;
    }

    return newInterceptor;
  }

  List<Input> _makeInputs(List<ParsedInput> parsed) {
    List<Input> inputs = <Input>[];

    for (ParsedInput input in parsed) {
      if (input is ParsedInputInterceptor) {
        inputs.add(new InputInterceptor(input.id, input.resultFrom.name));
      } else if (input is ParsedInputCookie) {
        inputs.add(new InputCookie(input.key));
      } else if (input is ParsedInputCookies) {
        inputs.add(new InputCookies());
      } else if (input is ParsedInputHeader) {
        inputs.add(new InputHeader(input.key));
      } else if (input is ParsedInputHeaders) {
        inputs.add(new InputHeaders());
      } else if (input is ParsedInputPathParams) {
        inputs.add(new InputPathParams(input.type.displayName));
      } else if (input is ParsedInputQueryParams) {
        inputs.add(new InputQueryParams(input.type.displayName));
      } else if (input is ParsedInputRouteResponse) {
        inputs.add(new InputRouteResponse());
      }
    }

    return inputs;
  }
}

String _getStringTo(ParameterElementWrap param) {
  if (!param.type.isBuiltin) {
    throw new Exception("Can only convert builtin types!");
  }

  String ret = "stringTo";

  if (param.type.isInt) {
    ret += "Int";
  } else if (param.type.isDouble) {
    ret += "Double";
  } else if (param.type.isNum) {
    ret += "Num";
  } else if (param.type.isBool) {
    ret += "Bool";
  } else if (param.type.isString) {
    ret = "";
  } else {
    throw new Exception("Can only convert builtin types!");
  }

  return ret;
}
